'''
Chet Gnegy
www.chetgnegy.com
Free Photo Collage Generator
GPL License

'''


from PIL import Image
import ExifTags
import glob, os
import operator
import random
import sys, getopt



N = 10

bestPicks = {}


def topN(arr, N):
  sorted_arr = sorted(arr.iteritems(), key=operator.itemgetter(1))
  return sorted_arr[0:N]

# Returns the elementwise average color
def getAvgColor(im):
  R_sum = 0
  G_sum = 0
  B_sum = 0
  for i in range(0, im.size[0]):
    for j in range(0, im.size[1]):
      (R,G,B) = im.getpixel((i,j))
      R_sum += R
      G_sum += G
      B_sum += B
  R_sum /= 128*128
  G_sum /= 128*128
  B_sum /= 128*128
  return (R_sum, G_sum, B_sum)

def quantize(trip,n):
  R = (trip[0]>>n)<<n
  G = (trip[1]>>n)<<n
  B = (trip[2]>>n)<<n
  return (R, G, B)


# Gets the most frequently used color in the image
def getFrequent(im):
  cols = {}
  for i in range(0, im.size[0]):
    for j in range(0, im.size[1]):
      pix = im.getpixel((i,j))
      pix = quantize(pix,3);
      cols[pix] = cols.get(pix,0) + 1
  freq = max(cols.iteritems(), key=operator.itemgetter(1))[0]
  return freq

# A metric for determining match quality
def matchQuality(match, target):
  match = quantize(match,3)
  R = match[0]
  G = match[1]
  B = match[2]
  dR = target[0] - R
  dG = target[1] - G
  dB = target[2] - B
  return  abs(dR) + abs(dG) + abs(dB)



def main(argv):
  #Handle input arguments
  thumb_size = 90
  target_tiles = 100
  blend_amt = 0.5
  inputfile = ''
  outputfile = 'outfile.jpeg'
  try:
    opts, args = getopt.getopt(argv,"hi:o:d:t:n:b:",["ifile=","ofile=","dir","tilesize=","numtiles", "blend_amt"])
  except getopt.GetoptError:
    print 'collage.py -i <inputfile> -d <directory> -o <outputfile> -t <tilesize> -n <numtiles> -b <blend_amt>'
    sys.exit(2)
  for opt, arg in opts:
    if opt == '-h':
      print 'collage.py -i <inputfile> -d <directory> -o <outputfile> -t <tilesize> -n <numtiles> -b <blend_amt>'
      sys.exit()
    elif opt in ("-i", "--ifile"):
      inputfile = arg
    elif opt in ("-o", "--ofile"):
      outputfile = arg
    elif opt in ("-d", "--dir"):
      directory = arg
    elif opt in ("-t", "--tilesize"):
      thumb_size = int(arg)
    elif opt in ("-n", "--numtiles"):
      target_tiles = int(arg)
    elif opt in ("-b", "--blend_amt"):
      blend_amt = float(arg)
  
  if directory[-1] == '/':
    directory = directory[0:-1]
  

   
  # Creates thumbnails for each image in the source set
  if not os.path.exists("source/"):
    os.makedirs("source/")

  print "Making thumbnails..."
  for infile in glob.glob(directory+"/*"):
    try:
      file, ext = os.path.splitext(infile)

      #Strip off the directory
      if "/" in file:
        file = file[file.rfind("/")+1:] 
      
      # We don't wanna make anything that is already there
      if os.path.isfile("source/"+file + "_thumb.jpeg") or "_thumb" in file:
        continue
      
      # Only opens image files
      try:
        im = Image.open(infile)
      except:
        continue

      # Rotates the thumbnail to the correct orientation
      try :
        for orientation in ExifTags.TAGS.keys() : 
          if ExifTags.TAGS[orientation]=='Orientation' : break 
        exif=dict(im._getexif().items())

        if   exif[orientation] == 3 : 
          im=im.rotate(180, expand=True)
        elif exif[orientation] == 6 : 
          im=im.rotate(270, expand=True)
        elif exif[orientation] == 8 : 
          im=im.rotate(90, expand=True)
      except:
        pass

      # Square image from the center
      min_dimension = min(im.size)
      if (min_dimension == im.size[0]):
        x = 0
        y = im.size[1]/2-min_dimension/2
      else:
        x = im.size[0]/2-min_dimension/2
        y = 0
      im = im.crop((x,y,x+min_dimension,y+min_dimension))
      
      im.thumbnail((thumb_size,thumb_size), Image.ANTIALIAS)
      im.save("source/" + file + "_thumb.jpeg", "JPEG")
    except: 
      print "Error reading file: ", file, ext

  thumbs_lib = {}
  thumbs_lib_avg = {}
  thumbs_lib_freq = {}
  found_some = False
  # Collects data on the individual images
  print "Collecting data..."
  for infile in glob.glob("source/*_thumb.jpeg"):
    file, ext = os.path.splitext(infile)
    try:
      im = Image.open(infile)
        
      print file
      thumbs_lib[file] = im
      thumbs_lib_avg[file] = getAvgColor(im)
      thumbs_lib_freq[file] = getFrequent(im)
      found_some = True
    except:
      print "Error reading file: ", file, ext

  if found_some == False:
    print "There are no thumbnails in this directory!\n"
    return

  print "Resizing target..."
  # Makes a small copy of the target file (with filtering)
  target = Image.open(inputfile)
  [x, y] = target.size
  x_ = x/(1.0*min(x,y))
  y_ = y/(1.0*min(x,y))
  x_tiles = int( x_*target_tiles )
  y_tiles = int( y_*target_tiles )
  target_thumb = target.copy()
  target_thumb.thumbnail((x_tiles,y_tiles), Image.ANTIALIAS)
  
  last_choose = ""
  # Creates a blank image that needs to be filled by the thumbnails
  final = Image.new("RGB", (thumb_size*target_thumb.size[0], thumb_size*target_thumb.size[1]), "white")
  print "Making collage..."
  for i in range(0, target_thumb.size[0]):
    for j in range(0, target_thumb.size[1]):
      target_pix = target_thumb.getpixel((i,j))
      
      for key in thumbs_lib_avg:
        metric1 = matchQuality(thumbs_lib_freq[key], target_pix)
        metric2 = matchQuality(thumbs_lib_avg[key], target_pix)
        metric = min(metric1,metric2)+(metric1+metric2)/2
        bestPicks[key] = metric
      
      N = 10
      choose = topN(bestPicks,N)
      k = random.randint(0, N-1)
      
      while (last_choose == choose[k][0]):
        k = random.randint(0, N-1)
      
      last_choose = choose[k][0]

      final.paste( thumbs_lib[last_choose], (
                  i*thumb_size, 
                  j*thumb_size, 
                  (i+1)*thumb_size, 
                  (j+1)*thumb_size) )
  print "Saving..."
  bigger = target_thumb.copy()
  bigger = bigger.resize(final.size, Image.NEAREST)
  final = Image.blend(final, bigger, blend_amt)
  final.save(outputfile, "JPEG")

if __name__ == "__main__":
   main(sys.argv[1:])